// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl.html       *
// * Copyright (C) 2008-2011 ZenJu (zhnmju123 AT gmx.de)                    *
// **************************************************************************
//
#include "unittest.h"
#include "../process_xml.h"
#include "../../shared/xml_base.h"
#include <wx/filefn.h>
#include "../../shared/standard_paths.h"

using namespace ffs3;


TEST(XmlInputOutput, GUI)
{
    //cleanup
    wxString testXml = wxT("test語.xml");
    if (wxFileExists(testXml))
        wxRemoveFile(testXml);

    ASSERT_FALSE(wxFileExists(testXml));

    //load structure with gui settings "guiCfg"
    xmlAccess::XmlGuiConfig guiCfg;
    guiCfg.mainCfg.compareVar = CMP_BY_TIME_SIZE;
    guiCfg.mainCfg.syncConfiguration.automatic       = false;
    guiCfg.mainCfg.syncConfiguration.exLeftSideOnly  = ffs3::SYNC_DIR_LEFT;
    guiCfg.mainCfg.syncConfiguration.exRightSideOnly = ffs3::SYNC_DIR_NONE;
    guiCfg.mainCfg.syncConfiguration.leftNewer       = ffs3::SYNC_DIR_LEFT;
    guiCfg.mainCfg.syncConfiguration.rightNewer      = ffs3::SYNC_DIR_RIGHT;
    guiCfg.mainCfg.syncConfiguration.different       = ffs3::SYNC_DIR_RIGHT;
    guiCfg.mainCfg.syncConfiguration.conflict        = ffs3::SYNC_DIR_RIGHT;
    guiCfg.mainCfg.globalFilter.includeFilter = Zstr("include語\\");
    guiCfg.mainCfg.globalFilter.excludeFilter = Zstr("語exclude\\8");
    guiCfg.mainCfg.handleDeletion = ffs3::MOVE_TO_CUSTOM_DIRECTORY;
    guiCfg.ignoreErrors  = false;

    //attention: hidden settings currently are NOT persisted!
    //guiCfg.mainCfg.hidden

    guiCfg.mainCfg.firstPair =
        FolderPairEnh(Zstr("&left1語ojk"),
                      Zstr("&right1語iu"),
                      boost::shared_ptr<AlternateSyncConfig>(),
                      FilterConfig());


    //add. pairs
    AlternateSyncConfig altCfg;
    altCfg.syncConfiguration.exLeftSideOnly = SYNC_DIR_LEFT;
    altCfg.handleDeletion = DELETE_PERMANENTLY;
    altCfg.customDeletionDirectory = wxT("239d i90q dasf 9");

    guiCfg.mainCfg.additionalPairs.push_back(
        FolderPairEnh(Zstr("&left1語"),
                      Zstr("&right1語"),
                      boost::shared_ptr<AlternateSyncConfig>(new AlternateSyncConfig(altCfg)),
                      FilterConfig()));

    FilterConfig altFilt;
    altFilt.includeFilter = Zstr("sdafi93423");
    altFilt.excludeFilter = Zstr("420opysdf");

    guiCfg.mainCfg.additionalPairs.push_back(
        FolderPairEnh(Zstr("国語1"),
                      Zstr("国語2"),
                      boost::shared_ptr<AlternateSyncConfig>(),
                      FilterConfig(altFilt)));


    guiCfg.hideFilteredElements = false;
    guiCfg.ignoreErrors         = true;
    guiCfg.syncPreviewEnabled   = false;

    //write XML
    ASSERT_NO_THROW(writeConfig(guiCfg, testXml));

    ASSERT_TRUE(xmlAccess::getXmlType(testXml) == xmlAccess::XML_GUI_CONFIG);

    //read XML
    xmlAccess::XmlGuiConfig guiCfg2;
    ASSERT_NO_THROW(xmlAccess::readConfig(testXml, guiCfg2));

    //test if retrieved file is equal created file
    ASSERT_TRUE(guiCfg == guiCfg2);

    //clean up
    wxRemoveFile(testXml);
}


TEST(XmlInputOutput, Batch)
{
    //cleanup
    wxString testXml = wxT("test語.xml");
    if (wxFileExists(testXml))
        wxRemoveFile(testXml);

    ASSERT_FALSE(wxFileExists(testXml));

    //load structure with batch settings "batchCfg"
    xmlAccess::XmlBatchConfig batchCfg;
    batchCfg.mainCfg.compareVar = CMP_BY_TIME_SIZE;
    batchCfg.mainCfg.syncConfiguration.automatic       = false;
    batchCfg.mainCfg.syncConfiguration.exLeftSideOnly  = ffs3::SYNC_DIR_LEFT;
    batchCfg.mainCfg.syncConfiguration.exRightSideOnly = ffs3::SYNC_DIR_NONE;
    batchCfg.mainCfg.syncConfiguration.leftNewer       = ffs3::SYNC_DIR_LEFT;
    batchCfg.mainCfg.syncConfiguration.rightNewer      = ffs3::SYNC_DIR_RIGHT;
    batchCfg.mainCfg.syncConfiguration.different       = ffs3::SYNC_DIR_RIGHT;
    batchCfg.mainCfg.globalFilter.includeFilter = Zstr("include語\\");
    batchCfg.mainCfg.globalFilter.excludeFilter = Zstr("語exclude\\8");
    batchCfg.mainCfg.handleDeletion = ffs3::MOVE_TO_CUSTOM_DIRECTORY;

    batchCfg.handleError = xmlAccess::ON_ERROR_IGNORE;
    batchCfg.mainCfg.firstPair =
        FolderPairEnh(Zstr("&left1語"),
                      Zstr("&right1語"),
                      boost::shared_ptr<AlternateSyncConfig>(),
                      FilterConfig());

    batchCfg.silent = true;

    batchCfg.logFileDirectory = wxT("custom deletion dir");

    //write XML
    ASSERT_NO_THROW(writeConfig(batchCfg, testXml));

    ASSERT_TRUE(xmlAccess::getXmlType(testXml) == xmlAccess::XML_BATCH_CONFIG);

    //read XML
    xmlAccess::XmlBatchConfig batchCfg2;
    ASSERT_NO_THROW(xmlAccess::readConfig(testXml, batchCfg2));

    //test if retrieved file is equal created file
    ASSERT_TRUE(batchCfg.mainCfg == batchCfg2.mainCfg);
    ASSERT_EQ(batchCfg.handleError, batchCfg2.handleError);
    ASSERT_EQ(batchCfg.silent, batchCfg2.silent);
    ASSERT_EQ(batchCfg.logFileDirectory, batchCfg2.logFileDirectory);

    //clean up
    wxRemoveFile(testXml);
}


inline
bool colAttrIsEqual(const xmlAccess::ColumnAttrib& a, const xmlAccess::ColumnAttrib& b)
{
    return (a.type     == b.type     &&
            a.visible  == b.visible  &&
            a.position == b.position &&
            a.width    == b.width);
}


TEST(XmlInputOutput, GlobalSettings)
{
    //load structure with settings
    xmlAccess::XmlGlobalSettings globalCfg;

    globalCfg.programLanguage         = wxLANGUAGE_ICELANDIC;
    globalCfg.optDialogs.warningDependentFolders = false;

    globalCfg.gui.widthNotMaximized    = 1111;
    globalCfg.gui.heightNotMaximized   = 2222;
    globalCfg.gui.posXNotMaximized     = 3333;
    globalCfg.gui.posYNotMaximized     = 4444;
    globalCfg.gui.isMaximized          = true;

    for (int i = 0; i < 4; ++i)
    {
        xmlAccess::ColumnAttrib newEntry;

        newEntry.type     = xmlAccess::ColumnTypes(i % 4);
        newEntry.visible  = true;
        newEntry.position = i;
        newEntry.width    = i + 1000;
        globalCfg.gui.columnAttribLeft.push_back(newEntry);

        newEntry.type     = xmlAccess::ColumnTypes(i % 4);
        newEntry.visible  = false;
        newEntry.position = i;
        newEntry.width    = i + 2000;
        globalCfg.gui.columnAttribRight.push_back(newEntry);
    }

    //write XML
    ASSERT_NO_THROW(writeConfig(globalCfg));

    ASSERT_TRUE(xmlAccess::getXmlType(xmlAccess::getGlobalConfigFile()) == xmlAccess::XML_GLOBAL_SETTINGS);

    //read XML
    xmlAccess::XmlGlobalSettings globalCfg2;
    ASSERT_NO_THROW(xmlAccess::readConfig(globalCfg2));

    //test if retrieved file is equal created file
    ASSERT_EQ(globalCfg.programLanguage, globalCfg2.programLanguage);
    ASSERT_EQ(globalCfg.optDialogs.warningDependentFolders, globalCfg2.optDialogs.warningDependentFolders);

    ASSERT_EQ(globalCfg.gui.widthNotMaximized, globalCfg2.gui.widthNotMaximized);
    ASSERT_EQ(globalCfg.gui.heightNotMaximized, globalCfg2.gui.heightNotMaximized);
    ASSERT_EQ(globalCfg.gui.posXNotMaximized, globalCfg2.gui.posXNotMaximized);
    ASSERT_EQ(globalCfg.gui.posYNotMaximized, globalCfg2.gui.posYNotMaximized);
    ASSERT_EQ(globalCfg.gui.isMaximized, globalCfg2.gui.isMaximized);
    ASSERT_EQ(globalCfg.gui.columnAttribLeft.size(), globalCfg2.gui.columnAttribLeft.size());
    ASSERT_EQ(globalCfg.gui.columnAttribRight.size(), globalCfg2.gui.columnAttribRight.size());
    ASSERT_TRUE(equal(globalCfg.gui.columnAttribLeft.begin(), globalCfg.gui.columnAttribLeft.end(), globalCfg2.gui.columnAttribLeft.begin(), colAttrIsEqual));
    ASSERT_TRUE(equal(globalCfg.gui.columnAttribRight.begin(), globalCfg.gui.columnAttribRight.end(), globalCfg2.gui.columnAttribRight.begin(), colAttrIsEqual));
}
