// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl.html       *
// * Copyright (C) 2008-2011 ZenJu (zhnmju123 AT gmx.de)                    *
// **************************************************************************
//
#include "../resources.h"
#include "../../shared/file_handling.h"
#include "../../shared/resolve_path.h"
#include <wx/sstream.h>
#include "unittest.h"
#include <wx/log.h>
#include "../../comparison.h"
#include "../../synchronization.h"
#include "../../algorithm.h"
#include "../../shared/global_func.h"
#include "../../shared/file_io.h"
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include "../../shared/string_conv.h"

using namespace ffs3;
using namespace boost::lambda;

unsigned int getTotalElementNoSub(const ffs3::HierarchyObject& hierObj)
{
    unsigned int total = 0;
    std::for_each(hierObj.useSubDirs().begin(), hierObj.useSubDirs().end(), boost::ref(total) += bind(getTotalElementNoSub, _1));

    return hierObj.useSubDirs().size() + hierObj.useSubFiles().size() + total;
}

unsigned int getTotalElementNo(const ffs3::FolderComparison& folderCmp)
{
    unsigned int total = 0;
    std::for_each(folderCmp.begin(), folderCmp.end(), boost::ref(total) += bind(getTotalElementNoSub, _1));
    return total;
}


std::vector<Zstring> getSampleNames(const unsigned int count)
{
    std::vector<Zstring> output;
    for (unsigned int i = 0; i < count; ++i)
    {
        Zstring newName;
        newName += Zstr('A') + i % 26;
        newName += Zstr('a') + (i * i) % 25;
        newName += Zstr('A') + (i + 3) % 24;
        newName += Zstr('A') + i % 11 + 10;
        //#ifdef FFS_WIN
        newName += wxToZ(wxT('語'));
        //#endif

        if (i % 2 != 0)
            newName += Zstr('.');
        newName += Zstr('a') + (i + count + 8) % 23;
        newName += Zstr('A') + (i + 13) % 22;
        newName += Zstr('a') + (i + count + 15) % 21;

        output.push_back(newName);
    }
    return output;
}


void syncByTimeDate(const Zstring& leftDir, const Zstring& rightDir, const SyncConfiguration& syncCfg, FolderComparison& output)
{
    output.clear();

    std::vector<FolderPairCfg> formattedFolders;
    FolderPairCfg newPair(
        ffs3::getFormattedDirectoryName(leftDir),
        ffs3::getFormattedDirectoryName(rightDir),
        BaseFilter::FilterRef(new NullFilter),
        syncCfg);
    formattedFolders.push_back(newPair);

    DummyStatusUpdater dummyStatus;

    //COMPARE DIRECTORIES
    xmlAccess::OptionalDialogs dummyWarnings;
    dummyWarnings.warningSignificantDifference = false;
    dummyWarnings.warningUnresolvedConflicts   = false;

    ffs3::CompareProcess comparison(ffs3::SYMLINK_IGNORE, 0, dummyWarnings, &dummyStatus);
    comparison.startCompareProcess(formattedFolders,
                                   CMP_BY_TIME_SIZE,
                                   output);

    //START SYNCHRONIZATION
    ffs3::SyncProcess synchronization(dummyWarnings, false, false, false, dummyStatus);

    std::vector<FolderPairSyncCfg> fpSyncCfg;
    fpSyncCfg.push_back(
        FolderPairSyncCfg(false,
                          ffs3::DELETE_PERMANENTLY,
                          Zstring()));

    synchronization.startSynchronizationProcess(fpSyncCfg, output);

    //all rows should be processed (with the exception of conflicts
    SyncStatistics finalStat(output);
    ASSERT_EQ(finalStat.getCreate() + finalStat.getOverwrite() + finalStat.getDelete(), 0);
    ASSERT_TRUE(finalStat.getDataToProcess() == 0);

    //compare again to get status after sync
    output.clear();
    //create new CompareProcess instance because of buffering!
    ffs3::CompareProcess comparison2(ffs3::SYMLINK_IGNORE, 0, dummyWarnings, &dummyStatus);
    comparison2.startCompareProcess(formattedFolders,
                                    CMP_BY_TIME_SIZE,
                                    output);
}


void createTestDirectories(const Zstring& leftDir, const Zstring& rightDir, std::vector<Zstring>& directoriesRelative, std::vector<Zstring>& filesRelative)
{
    //clean up
    ASSERT_NO_THROW(ffs3::removeDirectory(leftDir));
    ASSERT_NO_THROW(ffs3::removeDirectory(rightDir));

    ASSERT_FALSE(dirExists(leftDir));
    ASSERT_FALSE(dirExists(rightDir));

    //build vector of files and directories for testing
    directoriesRelative.clear();
    filesRelative.clear();

    directoriesRelative.push_back(Zstring());
    unsigned int totalNumberOfObjects = 0;
    for (int k = 1; k <= 3; ++k) //directory hierarchy to create
    {
        std::vector<Zstring> newNames = getSampleNames(10 / k);
        totalNumberOfObjects += newNames.size() * directoriesRelative.size() * 2;

        std::vector<Zstring> tmp;
        for (std::vector<Zstring>::iterator i = directoriesRelative.begin(); i != directoriesRelative.end(); ++i)
        {
            for (std::vector<Zstring>::iterator j = newNames.begin(); j != newNames.end(); ++j)
            {
                const Zstring newDir = *i + common::FILE_NAME_SEPARATOR + *j;
                tmp.push_back(newDir);

                const Zstring newFile = *i + common::FILE_NAME_SEPARATOR + Zstr("file") + *j;
                filesRelative.push_back(newFile);
            }
        }
        tmp.swap(directoriesRelative);
    }

    ASSERT_NO_THROW(
    {
        //create directories
        wxLogNull dummy;
        for (std::vector<Zstring>::iterator i = directoriesRelative.begin(); i != directoriesRelative.end(); ++i)
            ffs3::createDirectory(leftDir + *i);
    });

    ASSERT_NO_THROW(ffs3::createDirectory(rightDir));
    ASSERT_TRUE(dirExists(rightDir));

    //check that all directories on left have been created successfully
    ASSERT_TRUE(dirExists(leftDir));
    ASSERT_TRUE(dirExists(rightDir));
    for (std::vector<Zstring>::iterator i = directoriesRelative.begin(); i != directoriesRelative.end(); ++i)
        ASSERT_TRUE(dirExists(leftDir + *i));

    //create files
    for (std::vector<Zstring>::iterator i = filesRelative.begin(); i != filesRelative.end(); ++i)
    {
        FileOutput newFile(leftDir + *i, FileOutput::ACC_OVERWRITE);
        const char dummy[] = "dummy text";
        newFile.write(dummy, sizeof(dummy));
    }

    //check that all files have been created successfully
    for (std::vector<Zstring>::iterator i = filesRelative.begin(); i != filesRelative.end(); ++i)
        ASSERT_TRUE(fileExists(leftDir + *i));



    //fill right directory through sync "left->right"
    //###########################
    FolderComparison currentGridData;
    SyncConfiguration syncConfig;
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_RIGHT;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_NONE;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_NONE;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_NONE;
    syncConfig.different       = ffs3::SYNC_DIR_NONE;

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, currentGridData));

    //check if both folders contain the same files

    //make sure the correct number of files and directories was created
    ASSERT_EQ(getTotalElementNo(currentGridData), totalNumberOfObjects);
    ASSERT_TRUE(allElementsEqual(currentGridData));

    //check that all directories on right have been copied successfully
    for (std::vector<Zstring>::iterator i = directoriesRelative.begin(); i != directoriesRelative.end(); ++i)
        ASSERT_TRUE(dirExists(rightDir + *i));

    //check that all files have been created successfully on right
    for (std::vector<Zstring>::iterator i = filesRelative.begin(); i != filesRelative.end(); ++i)
        ASSERT_TRUE(fileExists(rightDir + *i));
}



void getDiffNumbersSub(const HierarchyObject& hierObj, int& equal, int& leftOnly, int& rightOnly, int& leftNewer, int& rightNewer, int& different, int& conflict)
{
    for (HierarchyObject::SubFileMapping::const_iterator i = hierObj.useSubFiles().begin(); i != hierObj.useSubFiles().end(); ++i)
        switch (i->getCategory())
        {
            case FILE_LEFT_SIDE_ONLY:
                leftOnly++;
                break;
            case FILE_RIGHT_SIDE_ONLY:
                rightOnly++;
                break;
            case FILE_RIGHT_NEWER:
                rightNewer++;
                break;
            case FILE_LEFT_NEWER:
                leftNewer++;
                break;
            case FILE_DIFFERENT:
                different++;
                break;
            case FILE_DIFFERENT_METADATA:
            case FILE_EQUAL:
                equal++;
                break;
            case FILE_CONFLICT:
                conflict++;
                break;
        }

    for (HierarchyObject::SubDirMapping::const_iterator i = hierObj.useSubDirs().begin(); i != hierObj.useSubDirs().end(); ++i)
    {
        //recursion
        getDiffNumbersSub(*i, equal, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);

        switch (i->getCategory())
        {
            case FILE_LEFT_SIDE_ONLY:
                leftOnly++;
                break;
            case FILE_RIGHT_SIDE_ONLY:
                rightOnly++;
                break;

            case FILE_EQUAL:
            case FILE_DIFFERENT_METADATA:
                equal++;
                break;

            case FILE_RIGHT_NEWER:
            case FILE_LEFT_NEWER:
            case FILE_DIFFERENT:
            case FILE_CONFLICT:
            default:
                FAIL();
                break;
        }
    }
}


void getDiffNumbers(const FolderComparison& folderCmp, int& leftOnly, int& rightOnly, int& leftNewer, int& rightNewer, int& different, int& conflict)
{
    int equal  = 0;
    leftOnly   = 0;
    rightOnly  = 0;
    leftNewer  = 0;
    rightNewer = 0;
    different  = 0;
    conflict   = 0;

    std::for_each(folderCmp.begin(), folderCmp.end(),
                  bind(getDiffNumbersSub, _1,
                       boost::ref(equal),
                       boost::ref(leftOnly),
                       boost::ref(rightOnly),
                       boost::ref(leftNewer),
                       boost::ref(rightNewer),
                       boost::ref(different),
                       boost::ref(conflict)));

    ASSERT_EQ(int(getTotalElementNo(folderCmp)), leftOnly + rightOnly + leftNewer + rightNewer + different + equal + conflict);
}


//predicate-formatter for asserting that two dates are equal
testing::AssertionResult assertEqual(const char* a_expr, const char* b_expr, const wxDateTime& a, const wxDateTime& b)
{
    if (a == b)
        return testing::AssertionSuccess();

    testing::Message msg;
    msg << a_expr << " = " << a.Format().c_str() <<
        " and " << b_expr << " = " << b.Format().c_str() << " not equal!";

    return testing::AssertionFailure(msg);
}


//predicate-formatter for asserting that two long integers are equal
testing::AssertionResult assertEqual(const char* a_expr, const char* b_expr, const wxULongLong& a, const wxULongLong& b)
{
    if (a == b)
        return testing::AssertionSuccess();

    testing::Message msg;
    msg << a_expr << " = " << a.ToString().c_str() <<
        " and " << b_expr << " = " << b.ToString().c_str() << " not equal!";

    return testing::AssertionFailure(msg);
}


//predicate-formatter for asserting that two Zstrings are equal
testing::AssertionResult assertEqual(const char* a_expr, const char* b_expr, const Zstring& a, const Zstring& b)
{
    if (a == b)
        return testing::AssertionSuccess();

    testing::Message msg;
    msg << a_expr << " = " << a.c_str() <<
        " and " << b_expr << " = " << b.c_str() << " not equal!";

    return testing::AssertionFailure(msg);
}


//predicate-formatter for asserting that two wxStrings are equal
testing::AssertionResult assertEqual(const char* a_expr, const char* b_expr, const wxString& a, const wxString& b)
{
    if (a == b)
        return testing::AssertionSuccess();

    testing::Message msg;
    msg << a_expr << " = " << a.c_str() <<
        " and " << b_expr << " = " << b.c_str() << " not equal!";

    return testing::AssertionFailure(msg);
}


class SynchronizationTest : public testing::Test
{
protected:
    // Per-test-case set-up.
    // Called before the first test in this test case.
    static void SetUpTestCase()
    {
        leftDir  = Zstr("unit_test_left");
        rightDir = Zstr("unit_test_right");

        //create and fill directories on HD for testing
        ASSERT_NO_THROW(createTestDirectories(leftDir, rightDir, directoriesRel, filesRel));

        //create differences of all types in folder structure
        restoreFolderDiff();

        //retrieve comparison result through dummy sync
        FolderComparison cmpResult2;
        SyncConfiguration syncConfig2;
        syncConfig2.automatic       = false;
        syncConfig2.exLeftSideOnly  = ffs3::SYNC_DIR_NONE;
        syncConfig2.exRightSideOnly = ffs3::SYNC_DIR_NONE;
        syncConfig2.leftNewer       = ffs3::SYNC_DIR_NONE;
        syncConfig2.rightNewer      = ffs3::SYNC_DIR_NONE;
        syncConfig2.different       = ffs3::SYNC_DIR_NONE;

        ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig2, cmpResult2));

        //check that "compare" returns the right numbers of each type
        initialLineCount = getTotalElementNo(cmpResult2);
        getDiffNumbers(cmpResult2, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
        ASSERT_EQ(leftOnly, 2);
        ASSERT_EQ(rightOnly, 2);
        ASSERT_EQ(leftNewer, 1);
        ASSERT_EQ(rightNewer, 1);
        ASSERT_EQ(different, 0);
        ASSERT_EQ(conflict, 1);
    }

    // Per-test-case tear-down.
    // Called after the last test in this test case.
    static void TearDownTestCase()
    {
        //clean up
        ASSERT_NO_THROW(ffs3::removeDirectory(leftDir));
        ASSERT_NO_THROW(ffs3::removeDirectory(rightDir));

        ASSERT_FALSE(dirExists(leftDir));
        ASSERT_FALSE(dirExists(rightDir));
    }


    static void restoreFolderDiff() //create differences in folder structure of all types
    {
        //left side only
        if (fileExists(rightDir + filesRel[20]))
            ASSERT_NO_THROW(removeFile(rightDir + filesRel[20]));
        if (dirExists(rightDir + directoriesRel[20]))
            ASSERT_NO_THROW(ffs3::removeDirectory(rightDir + directoriesRel[20]));
        if (!fileExists(leftDir + filesRel[20])) //restore test files/folders
        {
            FileOutput newFile(leftDir + filesRel[20], FileOutput::ACC_CREATE_NEW);
            const char dummy[] = "adskfj48fm4r4rgrg";
            newFile.write(dummy, sizeof(dummy));
        }
        if (!dirExists(leftDir + directoriesRel[20]))
            ASSERT_NO_THROW(ffs3::createDirectory(leftDir + directoriesRel[20]));

        //right side only
        if (fileExists(leftDir + filesRel[10]))
            ASSERT_NO_THROW(removeFile(leftDir + filesRel[10]));
        if (dirExists(leftDir + directoriesRel[10]))
            ASSERT_NO_THROW(ffs3::removeDirectory(leftDir + directoriesRel[10]));

        if (!fileExists(rightDir + filesRel[10])) //restore test files/folders
        {
            FileOutput newFile(rightDir + filesRel[10], FileOutput::ACC_CREATE_NEW);
            const char dummy[] = "3333skfj484g4hh5555555";
            newFile.write(dummy, sizeof(dummy));
        }
        if (!dirExists(rightDir + directoriesRel[10]))
            ASSERT_NO_THROW(ffs3::createDirectory(rightDir + directoriesRel[10]));

        //left newer
        {
            wxDateTime modTime;
            wxFileName file(zToWx(leftDir + filesRel[30]));
            ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
            modTime.Add(wxDateSpan(0,0,0,1));
            ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
        }

        //right newer
        {
            wxDateTime modTime;
            wxFileName file(zToWx(rightDir + filesRel[40]));
            ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
            modTime.Add(wxDateSpan(0,0,0,1));
            ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
        }

        //conflict: same date different size
        {
            wxDateTime modTime;
            wxFileName file(zToWx(rightDir + filesRel[50]));
            ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
            {
                FileOutput newFile(rightDir + filesRel[50], FileOutput::ACC_OVERWRITE);
                const char dummy[] = "1337"; //"most likely" a different size
                newFile.write(dummy, sizeof(dummy));

                //wxFile newFile(rightDir + filesRel[50], wxFile::write_append);
                //newFile.Write(Zstr(""));
            }

            ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
        }
    }

    //shared resources
    static unsigned int initialLineCount; //nr of rows in compareResult-vector
    static Zstring leftDir;
    static Zstring rightDir;

    static std::vector<Zstring> directoriesRel; //relative names
    static std::vector<Zstring> filesRel;

    //temporal variables
    static SyncConfiguration syncConfig;
    static int leftOnly;
    static int rightOnly;
    static int leftNewer;
    static int rightNewer;
    static int different;
    static int conflict;
    static wxDateTime modTimeLeft;
    static wxDateTime modTimeRight;
    static wxULongLong fileSizeLeft;
    static wxULongLong fileSizeRight;
};

unsigned int SynchronizationTest::initialLineCount;
Zstring SynchronizationTest::leftDir;
Zstring SynchronizationTest::rightDir;
std::vector<Zstring> SynchronizationTest::directoriesRel;
std::vector<Zstring> SynchronizationTest::filesRel;
SyncConfiguration SynchronizationTest::syncConfig;
int SynchronizationTest::leftOnly;
int SynchronizationTest::rightOnly;
int SynchronizationTest::leftNewer;
int SynchronizationTest::rightNewer;
int SynchronizationTest::different;
int SynchronizationTest::conflict;
wxDateTime SynchronizationTest::modTimeLeft;
wxDateTime SynchronizationTest::modTimeRight;
wxULongLong SynchronizationTest::fileSizeLeft;
wxULongLong SynchronizationTest::fileSizeRight;


TEST_F(SynchronizationTest, LeftSideOnly)
{
    //left side only(1)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_RIGHT;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_NONE;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_NONE;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_NONE;
    syncConfig.different       = ffs3::SYNC_DIR_NONE;

    FolderComparison cmpResult;
    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    ASSERT_TRUE(fileExists(rightDir + filesRel[20]));  //new files/dirs created
    ASSERT_TRUE(dirExists(rightDir + directoriesRel[20]));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 0);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 1);
    ASSERT_EQ(rightNewer, 1);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);

    //restore test files/folders
    ASSERT_NO_THROW(removeFile(rightDir + filesRel[20]));
    ASSERT_NO_THROW(ffs3::removeDirectory(rightDir + directoriesRel[20]));


    //left side only(2)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_LEFT;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_NONE;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_NONE;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_NONE;
    syncConfig.different       = ffs3::SYNC_DIR_NONE;

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    ASSERT_FALSE(fileExists(leftDir + filesRel[20]));  //files/dirs deleted
    ASSERT_FALSE(dirExists(leftDir + directoriesRel[20]));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 0);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 1);
    ASSERT_EQ(rightNewer, 1);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount - 2);

    //restore test files/folders
    ASSERT_FALSE(fileExists(leftDir + filesRel[20]));
    {
        FileOutput newFile(leftDir + filesRel[20], FileOutput::ACC_CREATE_NEW);
        const char dummy[] = "adskfjödsa";
        newFile.write(dummy, sizeof(dummy));
    }
    ASSERT_NO_THROW(ffs3::createDirectory(leftDir + directoriesRel[20]));
}


TEST_F(SynchronizationTest, RightSideOnly)
{
    //right side only(1)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_NONE;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_RIGHT;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_NONE;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_NONE;
    syncConfig.different       = ffs3::SYNC_DIR_NONE;

    FolderComparison cmpResult;
    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    ASSERT_FALSE(fileExists(rightDir + filesRel[10]));  //files/dirs deleted
    ASSERT_FALSE(dirExists(rightDir + directoriesRel[10]));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 0);
    ASSERT_EQ(leftNewer, 1);
    ASSERT_EQ(rightNewer, 1);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount - 2); //one file and directory were deleted

    //restore test files/folders
    ASSERT_FALSE(fileExists(rightDir + filesRel[10]));
    {
        FileOutput newFile(rightDir + filesRel[10], FileOutput::ACC_CREATE_NEW);
        const char dummy[] = "adskfjödsa";
        newFile.write(dummy, sizeof(dummy));
    }
    ASSERT_NO_THROW(ffs3::createDirectory(rightDir + directoriesRel[10]));


    //right side only(2)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_NONE;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_LEFT;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_NONE;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_NONE;
    syncConfig.different       = ffs3::SYNC_DIR_NONE;

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    ASSERT_TRUE(fileExists(leftDir + filesRel[10]));  //files/dirs created
    ASSERT_TRUE(dirExists(leftDir + directoriesRel[10]));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 0);
    ASSERT_EQ(leftNewer, 1);
    ASSERT_EQ(rightNewer, 1);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);

    //restore test files/folders
    ASSERT_NO_THROW(removeFile(leftDir + filesRel[10]));
    ASSERT_NO_THROW(ffs3::removeDirectory(leftDir + directoriesRel[10]));
}


TEST_F(SynchronizationTest, LeftNewer)
{
    //left newer(1)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_NONE;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_NONE;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_RIGHT;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_NONE;
    syncConfig.different       = ffs3::SYNC_DIR_NONE;

    modTimeLeft = wxFileName(zToWx(leftDir + filesRel[30])).GetModificationTime();
    modTimeRight = wxFileName(zToWx(rightDir + filesRel[30])).GetModificationTime();
    ASSERT_FALSE(modTimeLeft == modTimeRight);

    FolderComparison cmpResult;
    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    modTimeRight = wxFileName(zToWx(rightDir + filesRel[30])).GetModificationTime();
    ASSERT_PRED_FORMAT2(assertEqual, modTimeLeft, modTimeRight); //right file should be overwritten => mod time from left file should match

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 1);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);

    //restore test files/folders
    {
        wxDateTime modTime;
        wxFileName file(zToWx(leftDir + filesRel[30]));
        ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
        modTime.Add(wxDateSpan(0,0,0,1));
        ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
    }


    //left newer(2)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_NONE;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_NONE;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_LEFT;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_NONE;
    syncConfig.different       = ffs3::SYNC_DIR_NONE;

    modTimeLeft = wxFileName(zToWx(leftDir + filesRel[30])).GetModificationTime();
    modTimeRight = wxFileName(zToWx(rightDir + filesRel[30])).GetModificationTime();
    ASSERT_FALSE(modTimeLeft == modTimeRight);

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    modTimeLeft = wxFileName(zToWx(leftDir + filesRel[30])).GetModificationTime();
    ASSERT_PRED_FORMAT2(assertEqual, modTimeLeft, modTimeRight); //left file should be overwritten => mod time from right file should match

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 1);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);

    //restore test files/folders
    {
        wxDateTime modTime;
        wxFileName file(zToWx(leftDir + filesRel[30]));
        ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
        modTime.Add(wxDateSpan(0,0,0,1));
        ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
    }
}


TEST_F(SynchronizationTest, RightNewer)
{
    //right newer(1)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_NONE;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_NONE;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_NONE;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_RIGHT;
    syncConfig.different       = ffs3::SYNC_DIR_NONE;

    modTimeLeft = wxFileName(zToWx(leftDir + filesRel[40])).GetModificationTime();
    modTimeRight = wxFileName(zToWx(rightDir + filesRel[40])).GetModificationTime();
    ASSERT_FALSE(modTimeLeft == modTimeRight);

    FolderComparison cmpResult;
    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    modTimeRight = wxFileName(zToWx(rightDir + filesRel[40])).GetModificationTime();
    ASSERT_PRED_FORMAT2(assertEqual, modTimeLeft, modTimeRight); //right file should be overwritten => mod time from left file should match

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 1);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);
    //restore test files/folders
    {
        wxDateTime modTime;
        wxFileName file(zToWx(rightDir + filesRel[40]));
        ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
        modTime.Add(wxDateSpan(0,0,0,1));
        ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
    }


    //right newer(2)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_NONE;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_NONE;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_NONE;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_LEFT;
    syncConfig.different       = ffs3::SYNC_DIR_NONE;

    modTimeLeft = wxFileName(zToWx(leftDir + filesRel[40])).GetModificationTime();
    modTimeRight = wxFileName(zToWx(rightDir + filesRel[40])).GetModificationTime();
    ASSERT_FALSE(modTimeLeft == modTimeRight);

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    modTimeLeft = wxFileName(zToWx(leftDir + filesRel[40])).GetModificationTime();
    ASSERT_PRED_FORMAT2(assertEqual, modTimeLeft, modTimeRight); //left file should be overwritten => mod time from right file should match

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 1);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);
    //restore test files/folders
    {
        wxDateTime modTime;
        wxFileName file(zToWx(rightDir + filesRel[40]));
        ASSERT_TRUE(file.GetTimes(NULL, &modTime, NULL)); //get modification time
        modTime.Add(wxDateSpan(0,0,0,1));
        ASSERT_TRUE(file.SetTimes(NULL, &modTime, NULL)); //get modification time
    }
}


TEST_F(SynchronizationTest, OneWay)
{
    //one way sync(1)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_RIGHT;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_RIGHT;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_RIGHT;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_RIGHT;
    syncConfig.different       = ffs3::SYNC_DIR_RIGHT;

    FolderComparison cmpResult;
    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 0);
    ASSERT_EQ(rightOnly, 0);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount - 2);
    ASSERT_FALSE(fileExists(rightDir + filesRel[10]));
    ASSERT_FALSE(dirExists(rightDir + directoriesRel[10]));
    //restore test files/folders
    restoreFolderDiff(); //a little bit less precise as specific file manipulation


    //one way sync(2)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_LEFT;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_LEFT;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_LEFT;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_LEFT;
    syncConfig.different       = ffs3::SYNC_DIR_LEFT;

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 0);
    ASSERT_EQ(rightOnly, 0);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount - 2);
    ASSERT_FALSE(fileExists(leftDir + filesRel[20]));
    ASSERT_FALSE(dirExists(leftDir + directoriesRel[20]));
    //restore test files/folders
    restoreFolderDiff(); //a little bit less precise as specific file manipulation
}


TEST_F(SynchronizationTest, TwoWay)
{
    //two way sync(1)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_RIGHT;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_LEFT;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_RIGHT;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_LEFT;
    syncConfig.different       = ffs3::SYNC_DIR_LEFT; //should be irrelevant

    FolderComparison cmpResult;
    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 0);
    ASSERT_EQ(rightOnly, 0);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount);
    //restore test files/folders
    restoreFolderDiff(); //a little bit less precise as specific file manipulation


    //two way sync(2)
    //--------------------------------------------------------
    syncConfig.automatic       = false;
    syncConfig.exLeftSideOnly  = ffs3::SYNC_DIR_LEFT;
    syncConfig.exRightSideOnly = ffs3::SYNC_DIR_RIGHT;
    syncConfig.leftNewer       = ffs3::SYNC_DIR_LEFT;
    syncConfig.rightNewer      = ffs3::SYNC_DIR_RIGHT;
    syncConfig.different       = ffs3::SYNC_DIR_LEFT; //should be irrelevant

    ASSERT_NO_THROW(syncByTimeDate(leftDir, rightDir, syncConfig, cmpResult));

    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 0);
    ASSERT_EQ(rightOnly, 0);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 0);
    ASSERT_EQ(conflict, 1);

    ASSERT_EQ(getTotalElementNo(cmpResult), initialLineCount - 4);
    ASSERT_FALSE(fileExists(rightDir + filesRel[10]));
    ASSERT_FALSE(dirExists(rightDir + directoriesRel[10]));
    ASSERT_FALSE(fileExists(leftDir + filesRel[20]));
    ASSERT_FALSE(dirExists(leftDir + directoriesRel[20]));
    //restore test files/folders
    restoreFolderDiff(); //a little bit less precise as specific file manipulation
}


TEST_F(SynchronizationTest, CompareByContent)
{
    const Zstring newFileRel = Zstring(common::FILE_NAME_SEPARATOR) + Zstr("abcdefg.dat");

    ASSERT_FALSE(fileExists(rightDir + newFileRel));
    ASSERT_FALSE(fileExists(leftDir + newFileRel));

    //create additional file pair with different contents
    {
        FileOutput newFile(rightDir + newFileRel, FileOutput::ACC_CREATE_NEW);
        const char dummy[] = "123456789";
        newFile.write(dummy, sizeof(dummy));
    }
    {
        FileOutput newFile(leftDir + newFileRel, FileOutput::ACC_CREATE_NEW);
        const char dummy[] = "123456799";
        newFile.write(dummy, sizeof(dummy));
    }

    //setup sample files

    std::vector<FolderPairCfg> formattedFolders;
    FolderPairCfg newPair(
        ffs3::getFormattedDirectoryName(leftDir),
        ffs3::getFormattedDirectoryName(rightDir),
        BaseFilter::FilterRef(new NullFilter),
        SyncConfiguration());
    formattedFolders.push_back(newPair);


    DummyStatusUpdater dummyStatus;

    //COMPARE DIRECTORIES (by content)
    xmlAccess::OptionalDialogs dummyWarnings;
    dummyWarnings.warningSyncDatabase = false;

    FolderComparison cmpResult;
    ffs3::CompareProcess comparison(ffs3::SYMLINK_IGNORE, 0, dummyWarnings, &dummyStatus);
    comparison.startCompareProcess(formattedFolders,
                                   CMP_BY_CONTENT,
                                   cmpResult);


    getDiffNumbers(cmpResult, leftOnly, rightOnly, leftNewer, rightNewer, different, conflict);
    ASSERT_EQ(leftOnly, 2);
    ASSERT_EQ(rightOnly, 2);
    ASSERT_EQ(leftNewer, 0);
    ASSERT_EQ(rightNewer, 0);
    ASSERT_EQ(different, 2);
    ASSERT_EQ(conflict, 0);

    ASSERT_NO_THROW(removeFile(rightDir + newFileRel));
    ASSERT_NO_THROW(removeFile(leftDir + newFileRel));
}

