// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl.html       *
// * Copyright (C) 2008-2011 ZenJu (zhnmju123 AT gmx.de)                    *
// **************************************************************************
//
#include "unittest.h"
#include "../resources.h"
#include "../../shared/file_handling.h"
#include "../../shared/resolve_path.h"
#include "../../algorithm.h"
#include "../../shared/global_func.h"
#include "../../shared/util.h"
#include "../filter.h"
#include "../../shared/string_conv.h"
#include "../../shared/i18n.h"

using namespace ffs3;



TEST(Miscellaneous, RemoveRowsFromVector)
{
    std::vector<int> test; //0 1 2 3 4 5 6
    for (int i = 0; i < 7; ++i)
        test.push_back(i);

    std::set<unsigned int> rowsToRemove;
    rowsToRemove.insert(-41); //invalid
    rowsToRemove.insert(1);
    rowsToRemove.insert(2);
    rowsToRemove.insert(5);
    rowsToRemove.insert(15);  //invalid
    rowsToRemove.insert(100); //invalid

    common::removeRowsFromVector(rowsToRemove, test);
    //0 3 4 6

    ASSERT_EQ(unsigned(4), test.size());
    ASSERT_EQ(0, test[0]);
    ASSERT_EQ(3, test[1]);
    ASSERT_EQ(4, test[2]);
    ASSERT_EQ(6, test[3]);
}


TEST(Formatting, DirectoryNames)
{
#ifdef FFS_WIN
    ASSERT_PRED_FORMAT2(assertEqual, zToWx(ffs3::getFormattedDirectoryName(Zstr("  c:\\test "))), wxT("c:\\test\\"));
    ASSERT_PRED_FORMAT2(assertEqual, zToWx(ffs3::getFormattedDirectoryName(Zstr("  c:\\test\\ "))), wxT("c:\\test\\"));
    ASSERT_PRED_FORMAT2(assertEqual, zToWx(ffs3::getFormattedDirectoryName(Zstr("  \\abc\\ "))), wxT("C:\\abc\\"));
    ASSERT_PRED_FORMAT2(assertEqual, zToWx(ffs3::getFormattedDirectoryName(Zstr("  \\a\\b\\c "))), wxT("C:\\a\\b\\c\\"));
    ASSERT_PRED_FORMAT2(assertEqual, zToWx(ffs3::getFormattedDirectoryName(Zstr("  \\a\\b\\c d  "))), wxT("C:\\a\\b\\c d\\"));
    ASSERT_PRED_FORMAT2(assertEqual, zToWx(ffs3::getFormattedDirectoryName(Zstr("  \\\\a b/b\\c d  "))), wxT("\\\\a b\\b\\c d\\"));
#endif
}


//predicate-formatter for asserting that two vector<Zstring> are equal
testing::AssertionResult assertEqual(const char* a_expr, const char* b_expr, const std::vector<Zstring>& a, const std::vector<Zstring>& b)
{
    if (a == b)
        return testing::AssertionSuccess();

    testing::Message msg;
    msg << a_expr << " = " << '\n';
    for (std::vector<Zstring>::const_iterator i = a.begin(); i != a.end(); ++i)
        msg << i->c_str() << '\n';

    msg << " and " << b_expr << " = " << '\n';
    for (std::vector<Zstring>::const_iterator i = b.begin(); i != b.end(); ++i)
        msg << i->c_str() << '\n';

    msg << " not equal!";

    return testing::AssertionFailure(msg);
}

std::vector<Zstring> compoundStringToFilter(const Zstring& filterString);

TEST(Formatting, FilterString)
{
    std::vector<Zstring> filterActual = compoundStringToFilter(Zstr("   A\n   b;; c;\n\n;\n d    ;E    f  \n;& ;g\\"));
    std::vector<Zstring> filterExpected;
    filterExpected.push_back(Zstr("A"));
    filterExpected.push_back(Zstr("b"));
    filterExpected.push_back(Zstr("c"));
    filterExpected.push_back(Zstr("d"));
    filterExpected.push_back(Zstr("E    f"));
    filterExpected.push_back(Zstr("&"));
    filterExpected.push_back(Zstr("g\\"));
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);

    filterActual = compoundStringToFilter(Zstr("   ; ;\n a"));
    filterExpected.clear();
    filterExpected.push_back(Zstr("a"));
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);

    filterActual = compoundStringToFilter(Zstr("   ; ;\n       ;\n\n"));
    filterExpected.clear();
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);
}


TEST(Formatting, FileSize)
{
    wxString sizeActual = ffs3::formatFilesizeToShortString(807.0);
    wxString sizeExpected = _("%x Bytes");
    sizeExpected.Replace(wxT("%x"), wxT("807"));
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = ffs3::formatFilesizeToShortString(315621376.0);
    sizeExpected = _("%x MB");
    sizeExpected.Replace(wxT("%x"), wxT("301"));
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = ffs3::formatFilesizeToShortString(48954.0);
    sizeExpected = _("%x kB");
    sizeExpected.Replace(wxT("%x"), wxT("47.8"));
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = ffs3::formatFilesizeToShortString(35032904.0);
    sizeExpected = _("%x MB");
    sizeExpected.Replace(wxT("%x"), wxT("33.4"));
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = ffs3::formatFilesizeToShortString(3084582912.0);
    sizeExpected = _("%x GB");
    sizeExpected.Replace(wxT("%x"), wxT("2.87"));
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = ffs3::formatFilesizeToShortString(1001.0);
    sizeExpected = _("%x kB");
    sizeExpected.Replace(wxT("%x"), wxT("0.98"));
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);
}


TEST(Filtering, Standard)
{
#ifdef FFS_WIN
    ffs3::NameFilter filterIncl(Zstr("*abc?5*2  ;    abc?3;   name\\; test\\subdir "), Zstr(""));
#elif defined FFS_LINUX
    ffs3::NameFilter filterIncl(Zstr("*abc?5*2  ;    abc?3;   name/; test/subdir "), Zstr(""));
#endif

    ASSERT_TRUE(filterIncl.passFileFilter( Zstr("zzzzabcz5zzzzz2")));
    ASSERT_FALSE(filterIncl.passFileFilter(Zstr("zzzzabczz5zzzz2")));
    ASSERT_FALSE(filterIncl.passFileFilter(Zstr("zzzzabcz5zzzzz")));

    ASSERT_TRUE(filterIncl.passFileFilter(Zstr("abcz3")));
    ASSERT_FALSE(filterIncl.passFileFilter(Zstr("abcz")));

    bool continueTrav = true;
    ASSERT_TRUE(filterIncl.passDirFilter(Zstr("name"), &continueTrav));

    ASSERT_FALSE(filterIncl.passDirFilter(Zstr("test"), &continueTrav));
    ASSERT_TRUE(continueTrav); //but continue traversing, because subdir might match filter
    //------------------------------------------------------------------------------------------

#ifdef FFS_WIN
    ffs3::NameFilter filterExcl(Zstr("*"), Zstr("asdf?5 ;;; name\\*"));
#elif defined FFS_LINUX
    ffs3::NameFilter filterExcl(Zstr("*"), Zstr("asdf?5 ;;; name/*"));
#endif

    ASSERT_FALSE(filterExcl.passFileFilter(Zstr("asdfz5")));
    ASSERT_TRUE(filterExcl.passFileFilter(Zstr("asdfz")));

    ASSERT_FALSE(filterExcl.passDirFilter(Zstr("name"), &continueTrav));
    ASSERT_FALSE(continueTrav);
}


TEST(Zstring, AfterBefore)
{
    ASSERT_EQ(Zstring(Zstr("a/dsfsadf/abc")).AfterLast(Zchar('/')), Zstring(Zstr("abc")));
    ASSERT_EQ(Zstring(Zstr("a/dsfsadf/abc")).BeforeLast(Zchar('/')), Zstring(Zstr("a/dsfsadf")));
    ASSERT_EQ(Zstring(Zstr("a/dsfsadf/abc")).AfterFirst(Zchar('/')), Zstring(Zstr("dsfsadf/abc")));
    ASSERT_EQ(Zstring(Zstr("a/dsfsadf/abc")).BeforeFirst(Zchar('/')), Zstring(Zstr("a")));
}


TEST(Zstring, ReverseFind)
{
    Zstring phrase(Zstr("----abc-"));
    const size_t targetPos = 4;

    ASSERT_EQ(phrase.rfind(Zstr("abc")),    targetPos);
    ASSERT_EQ(phrase.rfind(Zstr("abc"), 7), targetPos);

    ASSERT_EQ(phrase.rfind(Zstr("abc"), 3), size_t(Zstring::npos));
    ASSERT_EQ(phrase.rfind(Zstr("abc"), 4), targetPos);
    ASSERT_EQ(phrase.rfind(Zstr("abc"), 5), targetPos);
}


TEST(Zstring, Resize)
{
    Zstring testStr(Zstr("abcdefgh"));
    testStr.resize(3);
    ASSERT_EQ(testStr, Zstring(Zstr("abc")));

    Zstring testStr2(Zstr("abcdefgh"));
    testStr2.resize(10); //add 0-characters!
    ASSERT_EQ(testStr2, Zstring(Zstr("abcdefgh\0\0"), 10));

    Zstring testStr3(Zstr("abcdefgh"));
    testStr3.resize(3, Zchar('N'));
    ASSERT_EQ(testStr3, Zstring(Zstr("abc")));

    Zstring testStr4(Zstr("abcdefgh"));
    testStr4.resize(10, Zchar('N')); //add 'N' characters!
    ASSERT_EQ(testStr4, Zstring(Zstr("abcdefghNN")));
}
