// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl.html       *
// * Copyright (C) 2008-2011 ZenJu (zhnmju123 AT gmx.de)                    *
// **************************************************************************
//
#include <iostream>
#include <fstream>
#include <set>
#include <map>
#include <algorithm>
#include <iterator>
#include "../shared/zstring.h"

//reads master translation file and adapts target language file accordingly, leaving not translated texts empty (which leads to an NSIS syntax error
//thous ensures that all texts are actually translated


//sample translation line:
//      LangString Lang_Help ${LANG_ENGLISH} "Help" ;comment
//      \____ Prefix ______/  \_Identifier_/  \_ Localized text _/

struct TranslationLine
{
    Zstring prefix;
    Zstring localization;
};


struct IsCorresponding
{
    IsCorresponding(const Zstring& prefix) : prefix_(prefix) {}

    bool operator()(const TranslationLine& transline) const
    {
        return transline.prefix == prefix_;
    }

private:
    const Zstring& prefix_;
};


void loadFile(const Zstring&        filename,
              std::vector<Zstring>& header,
              Zstring&              identifier,
              std::vector<TranslationLine>& textBuffer) //throw Zstring()
{
    header.clear();
    identifier.clear();
    textBuffer.clear();

    std::ifstream inputFile(filename.c_str());
    if (!inputFile)
        throw Zstring("Error opening file for reading: ") + filename;

    const int TEMP_SIZE = 100000;
    char temp[TEMP_SIZE];

    bool processingHeader = true;

    while (inputFile.getline(temp, TEMP_SIZE))
    {
        //remove UTF-8 Byte Ordering Mark
        if (    temp[0] == static_cast<char>(0xEF) &&
                temp[1] == static_cast<char>(0xBB) &&
                temp[2] == static_cast<char>(0xBF))
        {
            ::memmove(temp, temp + 3, TEMP_SIZE - 3);
        }

        const Zstring newLine = Zstring(temp).Trim(false); //trim from left

        if (newLine.StartsWith("LangString"))
        {
            processingHeader = false; //header ends with first occurence of "LangString"

            const size_t loc = newLine.find("${");
            if (loc != Zstring::npos)
            {
                const size_t loc2 = newLine.find("}", loc + 2);
                if (loc2 != Zstring::npos)
                {
                    const Zstring newIdentifier = Zstring(newLine.c_str() + loc, loc2 - loc + 1);

                    //set or test language id
                    if (identifier.empty())
                        identifier = newIdentifier;
                    else
                    {
                        if (identifier != newIdentifier)
                            throw Zstring("Invalid file: different language identifiers: ") + filename;
                    }

                    TranslationLine newEntry;
                    newEntry.prefix       = Zstring(newLine.c_str(), loc);
                    newEntry.localization = Zstring(newLine.c_str() + loc2 + 1);
                    textBuffer.push_back(newEntry);
                    continue;
                }
            }

            throw Zstring("Invalid file: not all rows contain valid language identifier: ") + filename;
        }
        else if (processingHeader && !newLine.StartsWith(";")) //don't preserve comments in header
            header.push_back(newLine);
    }

    //consistency checks
    if (header.empty())
        throw Zstring("Invalid file: header is empty: ") + filename;

    if (identifier.empty())
        throw Zstring("Invalid file: language identifier not found: ") + filename;

    if (textBuffer.empty())
        throw Zstring("Invalid file: no translation relevant rows found: ") + filename;
}


int main(int args, char** argv)
{
    if (args != 3)
    {
        std::cout << "\nInvalid input parameters: Specify master language file, e.g. \"english.nsh\", as (%1) and changing language file (%2)\n" << std::endl;
        system("pause");
        return -1;
    }

    const Zstring masterFileName   = argv[1];
    const Zstring changingFileName = argv[2];

    if (masterFileName.AfterLast('\\') == changingFileName.AfterLast('\\')) //input == changing file
        return 0;

    //master file
    std::vector<Zstring> masterHeader;
    Zstring masterIdentifier;
    std::vector<TranslationLine> masterTextBuffer;

    //changing file
    std::vector<Zstring> changingHeader;
    Zstring changingIdentifier;
    std::vector<TranslationLine> changingTextBuffer;

    try
    {
        loadFile(masterFileName,
                 masterHeader,
                 masterIdentifier,
                 masterTextBuffer); //throw Zstring()

        loadFile(changingFileName,
                 changingHeader,
                 changingIdentifier,
                 changingTextBuffer); //throw Zstring()

        if (masterIdentifier == changingIdentifier)
            throw Zstring("Both files have the same language identifier!");
    }
    catch (const Zstring& errorMsg)
    {
        std::cout << errorMsg << std::endl;
        system("pause");
        return -1;
    }


    //update changing file
    std::ofstream output(changingFileName.c_str());

    //write header
    std::copy(changingHeader.begin(), changingHeader.end(), std::ostream_iterator<const char*>(output, "\n"));

    bool notTranslatedTextsExist = false;
    for (std::vector<TranslationLine>::const_iterator i = masterTextBuffer.begin(); i != masterTextBuffer.end(); ++i)
    {
        const Zstring prefix   = i->prefix;
        const Zstring original = i->localization; //"localization"-section of master file is our original text
        Zstring localization;

        const std::vector<TranslationLine>::const_iterator iter = std::find_if(changingTextBuffer.begin(), changingTextBuffer.end(),
                                                                               IsCorresponding(prefix));  //inefficient...
        if (iter != changingTextBuffer.end() && !Zstring(iter->localization).Trim(true).empty())
            localization = iter->localization;
        else
            notTranslatedTextsExist = true; //leave newEntry.localized empty!

        //write body
        output << ";source text -> " << original << "\n"; //show original text to help translators
        output << prefix << changingIdentifier << localization << "\n\n";
    }

    if (notTranslatedTextsExist)
        std::cout << "\nTexts to be translated are existing!!!" << std::endl << std::endl;
    else
        std::cout << "All OK!" << std::endl << std::endl;

    return 0;
}
