// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl.html       *
// * Copyright (C) 2008-2011 ZenJu (zhnmju123 AT gmx.de)                    *
// **************************************************************************
//
#include <iostream>
#include <fstream>
#include <wx/wx.h>
#include <wx/dcmemory.h>
#include <wx/image.h>
#include <wx/wfstream.h>
#include <wx/zipstrm.h>
#include <memory>
#include <algorithm>
#include <string>
#include <vector>


void makeWhiteTransparent(const wxColor exceptColor, wxImage& image)
{
    unsigned char* alphaData = image.GetAlpha();
    if (alphaData)
    {
        assert(exceptColor.Red() != 255);

        unsigned char exCol = exceptColor.Red(); //alpha value can be extracted from any one of (red/green/blue)
        unsigned char* imageStart = image.GetData();

        unsigned char* j = alphaData;
        const unsigned char* const rowEnd = j + image.GetWidth() * image.GetHeight();
        while (j != rowEnd)
        {
            const unsigned char* imagePixel = imageStart + (j - alphaData) * 3; //each pixel consists of three chars
            //exceptColor(red,green,blue) becomes fully opaque(255), while white(255,255,255) becomes transparent(0)
            *(j++) = ((255 - imagePixel[0]) * wxIMAGE_ALPHA_OPAQUE) / (255 - exCol);
        }
    }
}


wxSize getSizeNeeded(const wxString& text, wxFont& font)
{
    wxCoord width, height;
    wxMemoryDC dc;

    wxString textFormatted = text;
    textFormatted.Replace(wxT("&"), wxT(""), false); //remove accelerator
    dc.GetMultiLineTextExtent(textFormatted, &width, &height , NULL, &font);
    return wxSize(width, height);
}


wxBitmap createBitmapFromText(const wxString& text)
{
    wxBitmapButton defaultButton;

    wxFont currentFont( 18, 74, 90, 92, false, wxT("Tahoma") );
    wxColor textColor  = defaultButton.GetForegroundColour();

    wxSize sizeNeeded = getSizeNeeded(text, currentFont);
    wxBitmap newBitmap(sizeNeeded.GetWidth(), sizeNeeded.GetHeight());

    wxMemoryDC dc;
    dc.SelectObject(newBitmap);

    //set up white background
    dc.SetBackground(*wxWHITE_BRUSH);
    dc.Clear();

    //find position of accelerator
    int indexAccel = -1;
    size_t accelPos;
    wxString textLabelFormatted = text;
    if ((accelPos = text.find(wxT("&"))) != wxString::npos)
    {
        textLabelFormatted.Replace(wxT("&"), wxT(""), false); //remove accelerator
        indexAccel = accelPos;
    }

    dc.SetTextForeground(textColor);
    dc.SetTextBackground(*wxWHITE);
    dc.SetFont(currentFont);
    dc.DrawLabel(textLabelFormatted, wxNullBitmap, wxRect(0, 0, newBitmap.GetWidth(), newBitmap.GetHeight()), wxALIGN_LEFT | wxALIGN_CENTER_VERTICAL, indexAccel);

    dc.SelectObject(wxNullBitmap);

    //add alpha channel to image
    // wxImage finalImage(newBitmap.ConvertToImage());
    //finalImage.SetAlpha();

    //linearInterpolation(finalImage);

    //make white background transparent
    //makeWhiteTransparent(textColor, finalImage);

    //return wxBitmap(finalImage);
    return newBitmap;
}


//copy one image into another, allowing arbitrary overlapping! (pos may contain negative numbers)
void writeToImage(const wxImage& source, const wxPoint pos, wxImage& target)
{
    //determine startpositions in source and target image, as well as width and height to be copied
    wxPoint posSrc, posTrg;
    int width, height;

    //X-axis
    if (pos.x < 0)
    {
        posSrc.x = -pos.x;
        posTrg.x = 0;
        width = std::min(pos.x + source.GetWidth(), target.GetWidth());
    }
    else
    {
        posSrc.x = 0;
        posTrg.x = pos.x;
        width = std::min(target.GetWidth() - pos.x, source.GetWidth());
    }

    //Y-axis
    if (pos.y < 0)
    {
        posSrc.y = -pos.y;
        posTrg.y = 0;
        height = std::min(pos.y + source.GetHeight(), target.GetHeight());
    }
    else
    {
        posSrc.y = 0;
        posTrg.y = pos.y;
        height = std::min(target.GetHeight() - pos.y, source.GetHeight());
    }


    if (width > 0 && height > 0)
    {
        //copy source to target respecting overlapping parts
        const unsigned char*       sourcePtr    = source.GetData() + 3 * (posSrc.x + posSrc.y * source.GetWidth());
        const unsigned char* const sourcePtrEnd = source.GetData() + 3 * (posSrc.x + (posSrc.y + height) * source.GetWidth());
        unsigned char*             targetPtr    = target.GetData() + 3 * (posTrg.x + posTrg.y * target.GetWidth());

        while (sourcePtr < sourcePtrEnd)
        {
            memcpy(targetPtr, sourcePtr, 3 * width);
            sourcePtr += 3 * source.GetWidth();
            targetPtr += 3 * target.GetWidth();
        }

        //handle different cases concerning alpha channel
        if (source.HasAlpha())
        {
            if (!target.HasAlpha())
            {
                target.SetAlpha();
                unsigned char* alpha = target.GetAlpha();
                memset(alpha, wxIMAGE_ALPHA_OPAQUE, target.GetWidth() * target.GetHeight());
            }

            //copy alpha channel
            const unsigned char*       sourcePtr    = source.GetAlpha() + (posSrc.x + posSrc.y * source.GetWidth());
            const unsigned char* const sourcePtrEnd = source.GetAlpha() + (posSrc.x + (posSrc.y + height) * source.GetWidth());
            unsigned char*             targetPtr    = target.GetAlpha() + (posTrg.x + posTrg.y * target.GetWidth());

            while (sourcePtr < sourcePtrEnd)
            {
                memcpy(targetPtr, sourcePtr, width);
                sourcePtr += source.GetWidth();
                targetPtr += target.GetWidth();
            }
        }
        else if (target.HasAlpha())
        {
            unsigned char*             targetPtr    = target.GetAlpha() + (posTrg.x + posTrg.y * target.GetWidth());
            const unsigned char* const targetPtrEnd = target.GetAlpha() + (posTrg.x + (posTrg.y + height) * target.GetWidth());

            while (targetPtr < targetPtrEnd)
            {
                memset(targetPtr, wxIMAGE_ALPHA_OPAQUE, width);
                targetPtr += target.GetWidth();
            }
        }
    }
}


std::string formatRcVerStr(const std::string& version)
{
    //format version string
    std::vector<std::string> verDigits;
    std::string number;
    for (std::string::const_iterator i = version.begin(); i != version.end(); ++i)
    {
        if (*i == '.' || *i == ',')
        {
            if (!number.empty())
            {
                verDigits.push_back(number);
                number.clear();
            }
            else
                verDigits.push_back("0");
        }
        else
            number += *i;
    }

    if (!number.empty()) //handle last digit
        verDigits.push_back(number);


    verDigits.resize(4, "0"); //final format: 1,0,0,0
    return verDigits[0] + "," +
           verDigits[1] + "," +
           verDigits[2] + "," +
           verDigits[3];
}


int main(int args, char** argv)
{
    std::cout<<"Enter new version number, e.g. \"1.17\": ";

    std::string versionNumber;
    std::cin>>versionNumber;

    //################################################################################
    //write version number to a file
    std::ofstream versionFile("version.txt");
    versionFile<<versionNumber;

    //################################################################################
    //write C++ include for version number
    std::ofstream versionIncl("version.h");
    versionIncl<<"#include <wx/string.h>"<<std::endl;
    versionIncl<<std::endl;
    versionIncl<<"namespace ffs3"<<std::endl;
    versionIncl<<"{"<<std::endl;
    versionIncl<<"    const wxString currentVersion = wxT(\""<<versionNumber<<"\"); //internal linkage!"<<std::endl;
    versionIncl<<"}"<<std::endl;

    //################################################################################
    //write Windows resource file include
    std::ofstream resourceIncl("version.rc");
    resourceIncl << "#define VER_FREEFILESYNC     " << formatRcVerStr(versionNumber) << "\n";
    resourceIncl << "#define VER_FREEFILESYNC_STR \"" << versionNumber << "\\0\""<<"\n";;

    //################################################################################
    //create header with version information for NSIS installer
    std::ofstream versionNsis("version.nsh");
    versionNsis<<"Name \"FreeFileSync v"<<versionNumber<<"\""<<std::endl;
    versionNsis<<"OutFile \"FreeFileSync_v"<<versionNumber<<"_setup.exe\""<<std::endl;
    versionNsis<<"!define FFS_Version \"FreeFileSync v"<<versionNumber<<"\""<<std::endl;

    //################################################################################
    //    //replace first row in Readme.txt
    //    if (!wxRenameFile(wxT("../BUILD/Readme.txt"), wxT("../BUILD/Readme.txt.tmp")))
    //    {
    //        std::cout<<"error renaming readme.txt"<<std::endl;
    //        system("pause");
    //        return -1;
    //    }
    //
    //    std::ifstream readmeIn("../BUILD/Readme.txt.tmp");
    //    std::ofstream readmeOut("../BUILD/Readme.txt");
    //
    //    const int bufferSize = 100000;
    //    char temp[bufferSize];
    //
    //    int rowNumber = 1;
    //    while (readmeIn.getline(temp, bufferSize))
    //    {
    //        if (rowNumber == 1)
    //            readmeOut<<"FreeFileSync v"<<versionNumber<<std::endl;
    //        else
    //            readmeOut<<temp<<std::endl;
    //        ++rowNumber;
    //    }
    //
    //    readmeIn.close();
    //    wxRemoveFile(wxT("../BUILD/Readme.txt.tmp"));

    //################################################################################
    //create new version bitmap

    //activate support for .png files
    wxImage::AddHandler(new wxPNGHandler); //ownership passed

    const int spaceBetween = 0;
    wxBitmap bitmapText = createBitmapFromText(wxString(wxT("FreeFileSync v")) + wxString::FromUTF8(versionNumber.c_str()) + wxT("     "));
    wxBitmap bitmapFront = wxImage(wxT("FFS Logo.png"), wxBITMAP_TYPE_PNG);

    //calculate dimensions of new button
    const int height = std::max(bitmapFront.GetHeight(), bitmapText.GetHeight());
    const int width  = bitmapFront.GetWidth() + spaceBetween + bitmapText.GetWidth();

    //create a opaque image
    wxImage transparentImage(width, height, false);
    transparentImage.SetAlpha();
    unsigned char* alpha = transparentImage.GetAlpha();
    memset(alpha, wxIMAGE_ALPHA_OPAQUE, width * height); //init the C way

    //white background
    unsigned char* imgData = transparentImage.GetData();
    std::fill(imgData, imgData + width * height * 3, 255); //init the C++ way

    //wxDC::DrawLabel() unfortunately isn't working for transparent images on Linux, so we need to use custom image-concatenation
    if (bitmapFront.IsOk())
        writeToImage(wxImage(bitmapFront.ConvertToImage()),
                     wxPoint(0, (transparentImage.GetHeight() - bitmapFront.GetHeight()) / 2),
                     transparentImage);

    if (bitmapText.IsOk())
        writeToImage(wxImage(bitmapText.ConvertToImage()),
                     wxPoint(bitmapFront.GetWidth() + spaceBetween, (transparentImage.GetHeight() - bitmapText.GetHeight()) / 2),
                     transparentImage);

    transparentImage.SaveFile(wxT("Logo.png"), wxBITMAP_TYPE_PNG);
    //################################################################################

    //update zip archive with new logo
    {
        wxFileInputStream fileInput(wxT("Logo.png"));
        if (!fileInput.IsOk())
        {
            std::cout<<"error opening logo.png"<<std::endl;
            system("pause");
            return -1;
        }


        std::auto_ptr<wxFileInputStream> in(new wxFileInputStream(wxT("../BUILD/Resources.dat")));
        wxTempFileOutputStream out(wxT("../BUILD/Resources.dat"));
        if (!out.IsOk())
        {
            std::cout<<"error opening zip output"<<std::endl;
            system("pause");
            return -1;
        }

        wxZipInputStream inzip(*in);
        wxZipOutputStream outzip(out);

        std::auto_ptr<wxZipEntry> entry;

        // transfer any meta-data for the archive as a whole (the zip comment
        // in the case of zip)
        outzip.CopyArchiveMetaData(inzip);

        // call CopyEntry for each entry except those matching the pattern
        while (entry.reset(inzip.GetNextEntry()), entry.get() != NULL)
            if (entry->GetName() == wxString(wxT("logo.png")))
            {
                if (!outzip.PutNextEntry(wxT("logo.png")))
                    break;
                outzip<<fileInput;
            }
            else
            {
                if (!outzip.CopyEntry(entry.release(), inzip))
                    break;
            }

        // close the input stream by releasing the pointer to it, do this
        // before closing the output stream so that the file can be replaced
        in.reset();

        // you can check for success as follows
        bool success = inzip.Eof() && outzip.Close() && out.Commit();

        if (!success)
        {
            std::cout<<"error writing zip file \"Resources.dat\"! Might be corrupted!"<<std::endl;
            system("pause");
            return -1;
        }
    }

    wxRemoveFile(wxT("Logo.png"));


    std::cout<<"\nOperation successful!\n"<<std::endl;
    system("pause");
    return 0;
}
